from xperms.base import BasePerm
from xperms.managers import RelatedPermManager


class Perm(BasePerm):

    related_manager = RelatedPermManager()

    class Meta:
        base_manager_name = 'related_manager'



from django.apps import apps
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType

from xperms import get_perm_model
from xperms.utils import get_content_type


Perm = get_perm_model()


def get_perm_from_permission_codename(permission_codename):
    # get permission object based on django hard-coded permission codename
    app_label, model_action = permission_codename.split('.', 1)
    action, model_name = model_action.split('_')
    model = apps.get_model(app_label, model_name)

    return Perm.objects.get(
        content_type=get_content_type(model),
        codename=action,
        object_id=None,
        field_name=None,
    )


class PermBackend(ModelBackend):

    # authentication backend, mostly for compatibility with django admin site

    def has_perm(self, user_obj, perm, obj=None):
        if user_obj.is_superuser:
            return True
        try:
            perm_obj = get_perm_from_permission_codename(perm)
        except Perm.DoesNotExist:
            perm_obj = None
        return user_obj.perms.has_perm(perm_obj, obj)

    def has_module_perms(self, user_obj, app_label):
        if not user_obj.is_active:
            return False

        ctypes = ContentType.objects.filter(app_label=app_label)

        return Perm.objects.filter(users__in=[user_obj,], content_type__in=ctypes)



from django.contrib.contenttypes.models import ContentType
from xperms.conf import settings

from xperms import get_perm_model, enums


def get_content_type(obj):
    return ContentType.objects.get_for_model(obj)


def get_perm(perm, obj=None):
    perm_model = get_perm_model()

    if isinstance(perm, perm_model):
        return perm

    perm_kwargs = perm_model.get_perm_kwargs(perm, obj)

    try:
        perm = perm_model.objects.get(**perm_kwargs)
    except perm_model.DoesNotExist:
        if settings.PERM_AUTO_CREATE:
            # create the required perm if it doesn't exist and is supposed to be created
            perm = perm_model.objects.create(**perm_kwargs)
        else:
            # check if a wildcard permission exists instead
            perm_kwargs['codename'] = enums.PERM_CODENAME_WILDCARD
            perm = perm_model.objects.get(**perm_kwargs)

    return perm



from xperms import enums
from xperms.models import Perm

from .base import ArticleTestCase, ArticleUserPermTestCase, ArticleGroupPermTestCase
from .factories import ArticleFactory


class ObjectPermTestCaseMixin:

    def setUp(self):
        super().setUp()
        self.article = ArticleFactory()
        self.article2 = ArticleFactory()

    def _create_perm(self):
        return self._create_add_perm()

    def _create_add_perm(self):
        return Perm.objects.create(
            type=enums.PERM_TYPE_OBJECT,
            codename=enums.PERM_CODENAME_ADD,
            content_object=self.article,
        )

    def _create_delete_perm(self):
        return Perm.objects.create(
            type=enums.PERM_TYPE_OBJECT,
            codename=enums.PERM_CODENAME_DELETE,
            content_object=self.article,
        )

    def _create_wildcard_perm(self):
        return Perm.objects.create(
            type=enums.PERM_TYPE_OBJECT,
            codename=enums.PERM_CODENAME_WILDCARD,
            content_object=self.article,
        )


class ObjectPermTestCase(ObjectPermTestCaseMixin, ArticleTestCase):

    def test_perm_has_correct_type(self):
        perm = self._create_perm()
        self.assertTrue(perm.is_object_perm())


class ArticleUserObjectPermPermTestCase(ObjectPermTestCaseMixin, ArticleUserPermTestCase):

    def test_add_object_perm_by_perm(self):
        perm = self._create_perm()

        self.user.perms.add_perm(perm)

        self.assertTrue(self.user.perms.has_perm(perm))

    def test_add_object_perm_by_str(self):
        add_obj_perm = self._create_add_perm()

        self.user.perms.add_perm('object.articles.Article.add', obj=self.article)

        self.assertTrue(self.user.perms.has_perm(add_obj_perm))

    def test_fail_add_object_perm_non_existent_codename(self):
        self._create_perm()
        with self.assertRaises(Perm.DoesNotExist):
            self.user.perms.add_perm('object.articles.Article.delete', obj=self.article)

    def test_fail_add_object_perm_non_existent_object(self):
        self._create_perm()
        with self.assertRaises(Perm.DoesNotExist):
            self.user.perms.add_perm('object.articles.Article.add', obj=self.article2)

    def test_has_model_perm_from_wildcard(self):
        self._create_wildcard_perm()
        self.user.perms.add_perm('object.articles.Article.*', obj=self.article)

        self.assertTrue(self.user.perms.has_perm('object.articles.Article.whatever', self.article))


class ArticleGroupObjectPermPermTestCase(ObjectPermTestCaseMixin, ArticleGroupPermTestCase):

    def test_add_object_perm_by_perm(self):
        perm = self._create_perm()

        self.group.perms.add_perm(perm)

        self.assertTrue(self.group.perms.has_perm(perm))

    def test_add_object_perm_by_str(self):
        add_obj_perm = self._create_add_perm()

        self.group.perms.add_perm('object.articles.Article.add', obj=self.article)

        self.assertTrue(self.group.perms.has_perm(add_obj_perm))

        # test perm is correctly available to the user as well
        self.assertFalse(self.user.perms.has_perm(add_obj_perm))

        self.user.groups.add(self.group)

        self.assertTrue(self.user.perms.has_perm(add_obj_perm))

    def test_fail_add_object_perm_by_non_existent_codename(self):
        self._create_perm()
        with self.assertRaises(Perm.DoesNotExist):
            self.group.perms.add_perm('object.articles.Article.delete', obj=self.article)

    def test_fail_add_object_perm_by_non_existent_object(self):
        self._create_perm()
        with self.assertRaises(Perm.DoesNotExist):
            self.group.perms.add_perm('object.articles.Article.add', obj=self.article2)



from django.contrib import admin
from xperms.admin import PermModelAdmin

from articles.models import Article


@admin.register(Article)
class ArticleAdmin(PermModelAdmin):

    perms_per_instance = True



from xperms import enums
from xperms.models import Perm

from .base import ArticleTestCase, ArticleUserPermTestCase, ArticleGroupPermTestCase


class ModelPermTestCaseMixin:

    def _create_perm(self):
        return self._create_add_perm()

    def _create_add_perm(self):
        return Perm.objects.create(
            type=enums.PERM_TYPE_MODEL,
            codename='add',
            content_type=self._get_content_type()
        )

    def _create_delete_perm(self):
        return Perm.objects.create(
            type=enums.PERM_TYPE_MODEL,
            codename='delete',
            content_type=self._get_content_type()
        )

    def _create_wildcard_perm(self):
        return Perm.objects.create(
            type=enums.PERM_TYPE_MODEL,
            codename=enums.PERM_CODENAME_WILDCARD,
            content_type=self._get_content_type()
        )


class ModelPermTestCase(ModelPermTestCaseMixin, ArticleTestCase):

    def test_perm_has_correct_type(self):
        perm = self._create_perm()
        self.assertTrue(perm.is_model_perm())


class ArticleUserModelPermPermTestCase(ModelPermTestCaseMixin, ArticleUserPermTestCase):

    def test_add_model_perm_by_perm(self):
        perm = self._create_perm()

        self.user.perms.add_perm(perm)

        self.assertTrue(self.user.perms.has_perm(perm))

    def test_add_model_perm_by_str(self):
        add_perm = self._create_add_perm()

        self.user.perms.add_perm('model.articles.Article.add')

        self.assertTrue(self.user.perms.has_perm(add_perm))

    def test_fail_add_model_perm_by_non_existent_codename(self):
        self._create_perm()
        with self.assertRaises(Perm.DoesNotExist):
            self.user.perms.add_perm('model.articles.Article.delete')

    def test_fail_add_model_perm_by_non_existent_model(self):
        self._create_perm()
        with self.assertRaises(LookupError):
            self.user.perms.add_perm('model.articles.Bar.fap')

    def test_has_model_perm_from_wildcard(self):
        self._create_wildcard_perm()

        self.user.perms.add_perm('model.articles.Article.*')

        self.assertTrue(self.user.perms.has_perm('model.articles.Article.whatever'))


class ArticleGroupModelPermPermTestCase(ModelPermTestCaseMixin, ArticleGroupPermTestCase):

    def test_add_model_perm_by_perm(self):
        perm = self._create_perm()

        self.group.perms.add_perm(perm)

        self.assertTrue(self.group.perms.has_perm(perm))

    def test_add_model_perm_by_str(self):
        add_perm = self._create_add_perm()

        self.group.perms.add_perm('model.articles.Article.add')

        self.assertTrue(self.group.perms.has_perm(add_perm))

        # test perm is correctly available to the user as well
        self.assertFalse(self.user.perms.has_perm(add_perm))

        self.user.groups.add(self.group)

        self.assertTrue(self.user.perms.has_perm(add_perm))

    def test_fail_add_model_perm_non_existent_codename(self):
        self._create_perm()
        with self.assertRaises(Perm.DoesNotExist):
            self.group.perms.add_perm('model.articles.Article.delete')

    def test_fail_add_model_perm_non_existent_model(self):
        self._create_perm()
        with self.assertRaises(LookupError):
            self.group.perms.add_perm('model.articles.Bar.fap')
